import numpy as np

"""
Kareem, L.A., Iwalewa, T.M., and Marhoun, M.al-.,: "CNew explicit correlation for the compressibility factor of natural gas: linearized z-factor isotherms," 
Journal of Petroleum Exploration and Production Technology (2016)

https://link.springer.com/article/10.1007/s13202-015-0209-3
"""


def kareem(Pr=None, Tr=None):

    a1 = 0.317842
    a2 = 0.382216
    a3 = -7.768354
    a4 = 14.290531
    a5 = 0.000002
    a6 = -0.004693
    a7 = 0.096254
    a8 = 0.166720
    a9 = 0.966910
    a10 = 0.063069
    a11 = -1.966847
    a12 = 21.0581
    a13 = -27.0246
    a14 = 16.23
    a15 = 207.783
    a16 = -488.161
    a17 = 176.29
    a18 = 1.88453
    a19 = 3.05921

    t = 1 / Tr

    A = a1 * t * np.exp(a2 * (1 - t) ** 2) * Pr
    B = a3 * t + a4 * t ** 2 + a5 * t ** 6 * Pr ** 6
    C = a9 + a8 * t * Pr + a7 * t ** 2 * Pr ** 2 + a6 * t ** 3 * Pr ** 3
    D = a10 * t * np.exp(a11 * (1 - t) ** 2)
    """
    Notes: 
    When Pr = 3.0153, Tr = 1.6155, the calculated values for E, F, and G does not agree with the sample calculation
    presented in the original paper. 
    
    Python codes:        Paper sample calculation:
    E =  6.52959         E = 6.56232
    F =  -16.61537       F = -17.08860
    G =  3.77819         G = 3.80545
    
    After many checks and hand-calculation, I concluded that the author made some mistakes while writing up his sample
    calculation steps. 
    
    """
    E = a12 * t + a13 * t ** 2 + a14 * t ** 3
    F = a15 * t + a16 * t ** 2 + a17 * t ** 3
    G = a18 + a19 * t

    #E = 6.56232
    #F = -17.08860
    #G = 3.80547

    y = (D * Pr) / (
            (1 + A ** 2) / C - (A ** 2 * B) / C ** 3
    )

    #print('t = ', t)
    #print('A = ', A)
    #print('B = ', B)
    #print('C = ', C)
    #print('D = ', D)
    #print('E = ', E)
    #print('F = ', F)
    #print('G = ', G)
    #print('y = ', y)

    return (D * Pr * (1 + y + y ** 2 - y ** 3)) / ((D * Pr + E * y ** 2 - F * y ** G) * (1 - y) ** 3)

#print(kareem(Pr=3.0153, Tr=1.6155))


from scipy import optimize
import numpy as np
import matplotlib.pyplot as plt

from gascompressibility.z_correlation.DAK import DAK
from gascompressibility.z_correlation.hall_yarborough import hall_yarborough
from gascompressibility.z_correlation.londono import londono
from gascompressibility.z_correlation.kareem import kareem
from gascompressibility.pseudocritical import Piper
from gascompressibility.pseudocritical import Sutton


models = {
    'DAK': DAK,
    'hall_yarborough': hall_yarborough,
    'londono': londono,
    'kareem': kareem,
}

zmodels_ks = '["DAK", "hall_yarborough", "londono", "kareem"]'
pmodels_ks = '["Sutton", "Piper"]'


def _get_guess_constant():
    return 0.900000765321234598723486


def _get_z_model(model='DAK'):

    if model not in models.keys():
        raise KeyError(
            'Z-factor model "%s" is not implemented. Choose from the list of available models: %s' % (model, zmodels_ks)
        )

    return models[model]


def _calc_z_explicit_implicit_helper(Pr, Tr, zmodel_func, zmodel_str, guess, newton_kwargs, ps_props):

    maxiter = 1000

    # Explicit models
    if zmodel_str in ['kareem']:
        if guess != _get_guess_constant():
            raise KeyError('calc_z(model="%s") got an unexpected argument "guess"' % zmodel_str)
        if newton_kwargs is not None:
            raise KeyError('calc_z(model="%s") got an unexpected argument "newton_kwargs"' % zmodel_str)
        Z = zmodel_func(Pr=Pr, Tr=Tr)

    # Implicit models: they require iterative convergence
    else:
        if newton_kwargs is None:
            Z = optimize.newton(zmodel_func, guess, args=(Pr, Tr), maxiter=maxiter)
        else:
            # work around to set default maxiter = 1000, unless use specifies it
            if newton_kwargs.pop('maxiter', None) is None:
                Z = optimize.newton(zmodel_func, guess, args=(Pr, Tr), maxiter=1000, **newton_kwargs)
            else:
                Z = optimize.newton(zmodel_func, guess, args=(Pr, Tr), **newton_kwargs)

    return Z


def calc_z(sg=None, P=None, T=None, H2S=None, CO2=None, N2=None, Pr=None, Tr=None, pmodel='Piper', zmodel='DAK',
           guess=_get_guess_constant(), newton_kwargs=None, ps_props=False, ignore_conflict=False, **kwargs):

    z_model = _get_z_model(model=zmodel)

    # Pr and Tr are already provided:
    if Pr is not None and Tr is not None:

        # Explicit models
        Z = _calc_z_explicit_implicit_helper(Pr, Tr, z_model, zmodel, guess, newton_kwargs, ps_props)
        if ps_props is True:
            ps_props = {'z': Z, 'Pr': Pr, 'Tr': Tr}
            return ps_props
        else:
            return Z

    # Pr and Tr are NOT provided:
    if pmodel == 'Piper':
        pc_instance = Piper()
        Tr, Pr = pc_instance._initialize_Tr_and_Pr(sg=sg, P=P, T=T, Tr=Tr, Pr=Pr, H2S=H2S, CO2=CO2, N2=N2, ignore_conflict=ignore_conflict, **kwargs)
    elif pmodel == 'Sutton':
        if N2 is not None:
            raise KeyError('pmodel="Sutton" does not support N2 as input. Set N2=None')
        pc_instance = Sutton()
        Tr, Pr = pc_instance._initialize_Tr_and_Pr(sg=sg, P=P, T=T, Tr=Tr, Pr=Pr, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict, **kwargs)
    else:
        raise KeyError(
            'Pseudo-critical model "%s" is not implemented. Choose from the list of available models: %s' % (pmodel, pmodels_ks)
        )

    Z = _calc_z_explicit_implicit_helper(Pr, Tr, z_model, zmodel, guess, newton_kwargs, ps_props)

    if ps_props is True:
        ps_props = {'z': Z}
        ps_props.update(pc_instance.ps_props)
        ps_props['Tr'] = Tr
        ps_props['Pr'] = Pr
        return ps_props
    else:
        return Z

def quickstart():

    xmax = 8
    Prs = np.linspace(0, xmax, xmax * 10 + 1)
    Prs = np.array([round(Pr, 1) for Pr in Prs])

    Trs = np.array([1.05, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0])

    results = {Tr: {
        'Pr': np.array([]),
        'Z': np.array([])
    } for Tr in Trs}

    for Tr in Trs:
        for Pr in Prs:
            z = calc_z(Tr=Tr, Pr=Pr)
            results[Tr]['Z'] = np.append(results[Tr]['Z'], [z], axis=0)
            results[Tr]['Pr'] = np.append(results[Tr]['Pr'], [Pr], axis=0)

    label_fontsize = 12

    fig, ax = plt.subplots(figsize=(8, 5))
    for Tr in Trs:

        Zs = results[Tr]['Z']
        idx_min = np.where(Zs == min(Zs))

        p = ax.plot(Prs, Zs)
        if Tr == 1.05:
            t = ax.text(Prs[idx_min] - 0.5, min(Zs) - 0.005, '$T_{r}$ = 1.2', color=p[0].get_color())
            t.set_bbox(dict(facecolor='white', alpha=0.9, edgecolor='white', pad=1))
        else:
            t = ax.text(Prs[idx_min] - 0.2, min(Zs) - 0.005, Tr, color=p[0].get_color())
            t.set_bbox(dict(facecolor='white', alpha=0.9, edgecolor='white', pad=1))

    ax.set_xlim(0, xmax)
    ax.minorticks_on()
    ax.grid(alpha=0.5)
    ax.grid(visible=True, which='minor', alpha=0.1)
    ax.spines.top.set_visible(False)
    ax.spines.right.set_visible(False)

    ax.set_ylabel('Compressibility Factor, $Z$', fontsize=label_fontsize)
    ax.set_xlabel('Pseudo-Reduced Pressure, $P_{r}$', fontsize=label_fontsize)
    ax.text(0.57, 0.08, '$T_{r}$ = Pseudo-Reduced Temperature', fontsize=11, transform=ax.transAxes,
            bbox=dict(facecolor='white'))

    def setbold(txt):
        return ' '.join([r"$\bf{" + item + "}$" for item in txt.split(' ')])

    ax.set_title(setbold('Real Gas Law Compressibility Factor - Z') + ", computed with GasCompressiblityFactor-py ",
                 fontsize=12, pad=10, x=0.445, y=1.06)
    ax.annotate('', xy=(-0.09, 1.05), xycoords='axes fraction', xytext=(1.05, 1.05),
                arrowprops=dict(arrowstyle="-", color='k'))

    fig.tight_layout()

    return results, fig, ax



import numpy as np

"""
Hall, K.R., and Yarborough, L.: "A new equation of state for Z-factor calculations," Oil and Gas Journal (1973)

https://www.researchgate.net/publication/284299884_A_new_equation_of_state_for_Z-factor_calculations
"""


def hall_yarborough(z=None, Pr=None, Tr=None):

    t = 1 / Tr

    A1 = 0.06125 * t * np.exp(-1.2 * (1 - t) ** 2)
    A2 = 14.76 * t - 9.76 * t ** 2 + 4.58 * t ** 3
    A3 = 90.7 * t - 242.2 * t ** 2 + 42.4 * t ** 3
    A4 = 2.18 + 2.82 * t

    ((A1 * Pr) / z)

    return -A1 * Pr \
        + (((A1 * Pr) / z) + ((A1 * Pr) / z) ** 2 + ((A1 * Pr) / z) ** 3 - ((A1 * Pr) / z) ** 4)/(1 - ((A1 * Pr) / z)) ** 3 \
        - A2 * ((A1 * Pr) / z) ** 2 + A3 * ((A1 * Pr) / z) ** A4




import pathlib
from setuptools import setup, find_packages


# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE/"README.md").read_text()


def readme():
    with open('README.md') as f:
        return f.read().strip()


def classifiers():
    with open('classifiers.txt') as f:
        return f.read().strip().split('\n')


setup(
    name='gascompressibility',
    version='0.1.6',
    packages=find_packages(exclude=[
        "tutorials",
        "LICENSE",
        ".gitignore",
        "README.md",
        "misc",
        ".travis.yml",
        "notes.py",
        "testruns.ipynb",
        "papers",
        "docs",
        "docs_save_6_24",
    ]),
    description='GasCompressibility-py is a Python library for calculating the gas compressibility factor, Z, based on real gas law.',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=classifiers(),
    license='MIT',
    author='Eric Kim',
    author_email='aegis4048@gmail.com',
    install_requires=[
        'numpy>=1.14',
        'scipy>=1.5',
        'matplotlib>=3.2.1',
    ],
    url='https://github.com/aegis4048/GasCompressibiltiy-py/tree/main',
)


# python setup.py sdist bdist_wheel
# python -m twine upload --skip-existing dist/*

# python -m twine upload --skip-existing --repository testpypi dist/*
# python setup.py install --user
# pip uninstall gascompressibility
# pip install gascompressibility


import inspect

from gascompressibility.utilities.utilities import calc_Fahrenheit_to_Rankine
from gascompressibility.utilities.utilities import calc_psig_to_psia


class Sutton():

    """
    Class object to calculate pseudo-critical properties.

    Based on Sutton's specific gravity correlation [1]_ and Wichert & Aziz correction for acid gases (:math:`H_2S` and :math:`CO_2`) [2]_.

    """

    def __init__(self):

        self.sg = None
        """specific gravity (dimensionless)"""
        self.T_f = None
        """temperature (°F)"""
        self.T = None
        """temperature (°R)"""
        self.P_g = None
        """pressure (psig)"""
        self.P = None
        """pressure (psia)"""
        self.H2S = None
        """mole fraction of H2S (dimensionless)"""
        self.CO2 = None
        """mole fraction of CO2 (dimensionless)"""

        self.Tpc = None
        """pseudo-critical temperature, Tpc (°R)"""
        self.Ppc = None
        """pseudo-critical pressure, Ppc (psia)"""
        self.A = None
        self.B = None
        self.e_correction = None
        """temperature-correction factor for acid gases, ε (°R)"""
        self.Tpc_corrected = None
        """corrected pseudo-critical temperature, T'pc (°R)"""
        self.Ppc_corrected = None
        """corrected pseudo-critical pressure, P'pc (psia)"""
        self.Tr = None
        """pseudo-reduced temperature, Tr (°R)"""
        self.Pr = None
        """pseudo-reduced pressure, Pr (psia)"""

        self.ps_props = {
            'Tpc': None,
            'Ppc': None,
            'e_correction': None,
            'Tpc_corrected': None,
            'Ppc_corrected': None,
            'Tr': None,
            'Pr': None,
        }
        """dictionary of pseudo-critical properties."""

        self._first_caller_name = None
        self._first_caller_kwargs = {}
        self._first_caller_is_saved = False

    def __str__(self):
        return str(self.ps_props)

    def __repr__(self):
        description = '<gascompressibility.pseudocritical.Sutton> class object with the following calculated attributes:\n{'
        items = '\n   '.join('%s: %s' % (k, v) for k, v in self.ps_props.items())
        return description + '\n   ' +items + '\n}'

    """sum of the mole fractions of CO2 and H2S in a gas mixture"""
    def _calc_A(self, H2S=None, CO2=None):
        self._initialize_H2S(H2S)
        self._initialize_CO2(CO2)
        self.A = self.H2S + self.CO2
        return self.A

    """mole fraction of H2S in a gas mixture"""
    def _calc_B(self, H2S=None):
        self._initialize_H2S(H2S)
        self.B = self.H2S
        return self.B

    def calc_Tpc(self, sg=None):
        """
        Calculates pseudo-critical temperature, Tpc (°R)

        Parameters
        ----------
        sg : float
            specific gravity of gas (dimensionless)

        Returns
        -------
        float
            pseudo-critical temperature, Tpc (°R)
        """
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_sg(sg)
        self.Tpc = 169.2 + 349.5 * self.sg - 74.0 * self.sg ** 2
        self.ps_props['Tpc'] = self.Tpc
        return self.Tpc

    def calc_Ppc(self, sg=None):
        """
        Calculates pseudo-critical pressure, Ppc (psia)

        Parameters
        ----------
        sg : float
            specific gravity of gas (dimensionless)

        Returns
        -------
        float
            pseudo-critical pressure, Ppc (psia)
        """
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_sg(sg)
        self.Ppc = 756.8 - 131.07 * self.sg - 3.6 * self.sg ** 2
        self.ps_props['Ppc'] = self.Ppc
        return self.Ppc

    def calc_e_correction(self, H2S=None, CO2=None):
        """
        Calculates the temperature-correction factor for acid gases, ε (°R)

        Parameters
        ----------
        H2S : float
            mole fraction of H2S (dimensionless)
        CO2 : float
            mole fraction of CO2 (dimensionless)

        Returns
        -------
        float
            temperature-correction factor for acid gases, ε (°R)
        """
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_A(A=None, H2S=H2S, CO2=CO2)
        self._initialize_B(B=None, H2S=H2S)
        self.e_correction = 120 * (self.A ** 0.9 - self.A ** 1.6) + 15 * (self.B ** 0.5 - self.B ** 4)
        self.ps_props['e_correction'] = self.e_correction
        return self.e_correction

    def calc_Tpc_corrected(self, sg=None, Tpc=None, e_correction=None, H2S=None, CO2=None, ignore_conflict=False):
        """
        Calculates the corrected pseudo-critical temperature, T'pc (°R)

        Parameters
        ----------
        sg : float
            specific gravity of gas (dimensionless)
        Tpc : float
            pseudo-critical temperature, Tpc (°R)
        e_correction : float
            temperature-correction factor for acid gases, ε (°R)
        H2S : float
            mole fraction of H2S (dimensionless)
        CO2 : float
            mole fraction of CO2 (dimensionless)
        ignore_conflict : bool
            set this to True to override calculated variables with input keyword arguments.

        Returns
        -------
        float
            corrected pseudo-critical temperature, T'pc (°R)

        """
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_Tpc(Tpc, sg=sg, ignore_conflict=ignore_conflict)

        # Correction is not needed if no sour gas is present
        if e_correction is None and H2S is None and CO2 is None:
            self.Tpc_corrected = self.Tpc
            self.ps_props['Tpc_corrected'] = self.Tpc_corrected
            return self.Tpc_corrected

        self._initialize_e_correction(e_correction, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        self.Tpc_corrected = self.Tpc - self.e_correction
        self.ps_props['Tpc_corrected'] = self.Tpc_corrected
        return self.Tpc_corrected

    def calc_Ppc_corrected(self, sg=None, Tpc=None, Ppc=None, e_correction=None, Tpc_corrected=None, H2S=None, CO2=None, ignore_conflict=False):
        """
        Calculates the corrected pseudo-critical pressure, P'pc (psia)

        Parameters
        ----------
        sg : float
            specific gravity of gas (dimensionless)
        Tpc : float
            pseudo-critical temperature, Tpc (°R)
        Ppc : float
            pseudo-critical pressure, Ppc (psia)
        e_correction : float
            temperature-correction factor for acid gases, ε (°R)
        Tpc_corrected : float
            corrected pseudo-critical temperature, T'pc (°R)
        H2S : float
            mole fraction of H2S (dimensionless)
        CO2 : float
            mole fraction of CO2 (dimensionless)
        ignore_conflict : bool
            set this to True to override calculated variables with input keyword arguments.

        Returns
        -------
        float
            corrected pseudo-critical pressure, P'pc (psia)
        """
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_Ppc(Ppc, sg=sg, ignore_conflict=ignore_conflict)

        # Correction is not needed if no sour gas is present
        if e_correction is None and H2S is None and CO2 is None and Tpc is None and Tpc_corrected is None:
            self.Ppc_corrected = self.Ppc
            self.ps_props['Ppc_corrected'] = self.Ppc_corrected
            return self.Ppc_corrected

        self._initialize_Tpc(Tpc, sg=sg, ignore_conflict=ignore_conflict)
        self._initialize_B(B=None, H2S=H2S)
        self._initialize_e_correction(e_correction, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        self._initialize_Tpc_corrected(Tpc_corrected, sg=sg, Tpc=Tpc, e_correction=e_correction, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        self.Ppc_corrected = (self.Ppc * self.Tpc_corrected) / (self.Tpc - self.B * (1 - self.B) * self.e_correction)
        self.ps_props['Ppc_corrected'] = self.Ppc_corrected
        return self.Ppc_corrected

    def calc_Tr(self, T=None, Tpc_corrected=None, sg=None, Tpc=None, e_correction=None, H2S=None, CO2=None, ignore_conflict=False):
        """
        Calculates pseudo-reduced temperature, Tr (°R)

        Parameters
        ----------
        T : float
            temperature of gas (°F)
        Tpc_corrected : float
            corrected pseudo-critical temperature, T'pc (°R)
        sg : float
            specific gravity of gas (dimensionless)
        Tpc : float
            pseudo-critical temperature, Tpc (°R)
        e_correction : float
            temperature-correction factor for acid gases, ε (°R)
        H2S : float
            mole fraction of H2S (dimensionless)
        CO2 : float
            mole fraction of CO2 (dimensionless)
        ignore_conflict : bool
            set this to True to override calculated variables with input keyword arguments.

        Returns
        -------
        float
            pseudo-reduced temperature, Tr (°R)
        """
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_T(T)
        self._initialize_Tpc_corrected(Tpc_corrected, sg=sg, Tpc=Tpc, e_correction=e_correction, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        self.Tr = self.T / self.Tpc_corrected
        self.ps_props['Tr'] = self.Tr
        return self.Tr

    """pseudo-reduced pressure (psi)"""
    def calc_Pr(self, P=None, Ppc_corrected=None, sg=None, Tpc=None, Ppc=None, e_correction=None, Tpc_corrected=None, H2S=None, CO2=None, ignore_conflict=False):
        """
        Calculates pseudo-reduced pressure, Pr (psia)

        Parameters
        ----------
        P : float
            pressure of gas (psig)
        Ppc_corrected : float
            corrected pseudo-critical pressure, P'pc (psia)
        sg : float
            specific gravity of gas (dimensionless)
        Tpc : float
            pseudo-critical temperature, Tpc (°R)
        Ppc : float
            pseudo-critical pressure, Ppc (psia)
        e_correction : float
            temperature-correction factor for acid gases, ε (°R)
        Tpc_corrected : float
            corrected pseudo-critical temperature, T'pc (°R)
        H2S : float
            mole fraction of H2S (dimensionless)
        CO2 : float
            mole fraction of CO2 (dimensionless)
        ignore_conflict : bool
            set this to True to override calculated variables with input keyword arguments.

        Returns
        -------
        float
            pseudo-reduced pressure, Pr (psia)
        """
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_P(P)
        self._initialize_Ppc_corrected(Ppc_corrected, sg=sg, Tpc=Tpc, Ppc=Ppc, e_correction=e_correction, Tpc_corrected=Tpc_corrected, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        self.Pr = self.P / self.Ppc_corrected
        self.ps_props['Pr'] = self.Pr
        return self.Pr

    """This function is used by z_helper.py's calc_z function to check redundant arguments for Pr and Tr"""
    def _initialize_Tr_and_Pr(self, sg=None, P=None, T=None, Tpc=None, Ppc=None, Tpc_corrected=None, Ppc_corrected=None,
               H2S=None, CO2=None, Tr=None, Pr=None, e_correction=None, ignore_conflict=False):
        self._set_first_caller_attributes(inspect.stack()[0][3], locals())
        self._initialize_Tr(Tr, T, Tpc_corrected=Tpc_corrected, sg=sg, Tpc=Tpc, e_correction=e_correction, H2S=H2S,
                            CO2=CO2, ignore_conflict=ignore_conflict)
        self._initialize_Pr(Pr, P=P, Ppc_corrected=Ppc_corrected, sg=sg, Tpc=Tpc, Ppc=Ppc, e_correction=e_correction,
                            Tpc_corrected=Tpc_corrected, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        return self.Tr, self.Pr


    def _set_first_caller_attributes(self, func_name, func_kwargs):
        """
        Helper function to set properties related to the first function called (first in the call stack).
        This function doesn't do anything for 'calc_...()' functions called inside the first function.
        For exmaple, if `calc_Ppc_corrected()' is called, this function is skipped for 'calc_Ppc()' function which is
        triggered inside `calc_Ppc_corrected()`.
        :param func_name: string
            ex1) func_name= "calc_Tpc_corrected",
            ex2) func_name = "calc_Ppc_corrected",
        :param func_kwargs: kwarg parameters passed to 'func_name'
            ex1) func_kwargs = {'self': ...some_string, 'sg': None, 'Tpc': 377.59, 'e_correction': None, 'H2S': 0.07, 'CO2': 0.1}
            ex2) func_kwargs = {'self': ...some_string, 'sg': 0.6, 'Tpc': None, 'e_correction': None, 'H2S': 0.07, 'CO2': 0.1}
        """
        if not self._first_caller_is_saved:
            func_kwargs = {key: value for key, value in func_kwargs.items() if key != 'self'}
            if 'ignore_conflict' in func_kwargs:

                if func_kwargs['ignore_conflict'] is False:
                    """
                    This modification is needed for self._check_conflicting_arguments(). 
                    The exception in self._check_conflicting_arguments() compares if "kwarg is not None"
                    The default 'ignore_conflict' is a boolean object, so comparing if "ignore_conflict is not None"
                    will raise type error 
                    """
                    func_kwargs['ignore_conflict'] = None

            self._first_caller_name = func_name
            self._first_caller_kwargs = func_kwargs
            self._first_caller_is_saved = True
        else:
            pass

    def _check_conflicting_arguments(self, func, calculated_var):
        """
        :param func: string
            ex1) func_name = "calc_Tpc",
            ex2) func_name = "calc_J",
        :param calculated_var: string
            ex1) calculated_var = 'Tpc'
            ex1) calculated_var = 'J'
        """

        args = inspect.getfullargspec(func).args[1:]  # arg[0] = 'self', args = arguments defined in "func"
        for arg in args:
            if self._first_caller_kwargs[arg] is not None:
                # this is triggered in z_helper.py's calc_z() function
                if self._first_caller_name == '_initialize_Tr_and_Pr':
                    raise TypeError('%s() has conflicting keyword arguments "%s" and "%s"' % ('calc_z', calculated_var, arg))

                raise TypeError('%s() has conflicting keyword arguments "%s" and "%s"' % (self._first_caller_name, calculated_var, arg))

    def _initialize_sg(self, sg):
        if sg is None:
            if self._first_caller_name == 'calc_Ppc' or self._first_caller_name == 'calc_Tpc':
                raise TypeError("Missing a required argument, sg (specific gravity, dimensionless)")
            else:
                raise TypeError("Missing a required arguments, sg (specific gravity, dimensionless), or Tpc "
                                "(pseudo-critical temperature, °R) or Ppc (pseudo-critical pressure, psia). "
                                "Either both Tpc and Ppc must be inputted, or only sg needs to be inputted. "
                                "Both Tpc and Ppc can be computed from sg")
        else:
            self.sg = sg

    def _initialize_P(self, P):
        if P is None:
            raise TypeError("Missing a required argument, P (gas pressure, psig)")
        else:
            self.P_a = P  # psia
            self.P = calc_psig_to_psia(P)

    def _initialize_T(self, T):
        if T is None:
            raise TypeError("Missing a required argument, T (gas temperature, °F)")
        else:
            self.T_f = T
            self.T = calc_Fahrenheit_to_Rankine(T)

    def _initialize_H2S(self, H2S):
        if H2S is None:
            self.H2S = 0
        else:
            self.H2S = H2S

    def _initialize_CO2(self, CO2):
        if CO2 is None:
            self.CO2 = 0
        else:
            self.CO2 = CO2

    # The first argument A will always be None when called. However, still defining it for structural consistency
    def _initialize_A(self, A, H2S=None, CO2=None):
        if A is None:
            self._calc_A(H2S=H2S, CO2=CO2)
        else:
            self.A = A

    # The first argument B will always be None when called. However, still defining it for structural consistency
    def _initialize_B(self, B, H2S=None):
        if B is None:
            self._calc_B(H2S=H2S)
        else:
            self.B = B

    def _initialize_Tpc(self, Tpc, sg=None, ignore_conflict=None):
        if Tpc is None:
            self.calc_Tpc(sg=sg)
        else:
            if ignore_conflict is False:
                self._check_conflicting_arguments(self.calc_Tpc, 'Tpc')
            self.Tpc = Tpc

    def _initialize_Ppc(self, Ppc, sg=None, ignore_conflict=None):
        if Ppc is None:
            self.calc_Ppc(sg=sg)
        else:
            if ignore_conflict is False:
                self._check_conflicting_arguments(self.calc_Ppc, 'Ppc')
            self.Ppc = Ppc

    def _initialize_e_correction(self, e_correction, H2S=None, CO2=None, ignore_conflict=False):
        if e_correction is None:
            self.calc_e_correction(H2S=H2S, CO2=CO2)
        else:
            if ignore_conflict is False:
                self._check_conflicting_arguments(self.calc_e_correction, 'e_correction')
            self.e_correction = e_correction

    def _initialize_Tpc_corrected(self, Tpc_corrected, sg=None, Tpc=None, e_correction=None, H2S=None, CO2=None, ignore_conflict=False):
        if Tpc_corrected is None:
            self.calc_Tpc_corrected(sg=sg, Tpc=Tpc, e_correction=e_correction, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        else:
            if ignore_conflict is False:
                self._check_conflicting_arguments(self.calc_Tpc_corrected, 'Tpc_corrected')
            self.Tpc_corrected = Tpc_corrected

    def _initialize_Ppc_corrected(self, Ppc_corrected, sg=None, Tpc=None, Ppc=None, e_correction=None,
                                  Tpc_corrected=None, H2S=None, CO2=None, ignore_conflict=False):
        if Ppc_corrected is None:
            self.calc_Ppc_corrected(sg=sg, Tpc=Tpc, Ppc=Ppc, e_correction=e_correction, Tpc_corrected=Tpc_corrected, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        else:
            if ignore_conflict is False:
                self._check_conflicting_arguments(self.calc_Ppc_corrected, 'Ppc_corrected')
            self.Ppc_corrected = Ppc_corrected

    def _initialize_Pr(self, Pr, P=None, Ppc_corrected=None, sg=None, Tpc=None, Ppc=None, e_correction=None, Tpc_corrected=None, H2S=None, CO2=None, ignore_conflict=False):
        if Pr is None:
            self.calc_Pr(P=P, Ppc_corrected=Ppc_corrected, sg=sg, Tpc=Tpc, Ppc=Ppc, e_correction=e_correction, Tpc_corrected=Tpc_corrected, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        else:
            if ignore_conflict is False:
                self._check_conflicting_arguments(self.calc_Pr, 'Pr')
            self.Pr = Pr

    def _initialize_Tr(self, Tr, T, Tpc_corrected=None, sg=None, Tpc=None, e_correction=None, H2S=None, CO2=None, ignore_conflict=False):
        if Tr is None:
            self.calc_Tr(T=T, Tpc_corrected=Tpc_corrected, sg=sg, Tpc=Tpc, e_correction=e_correction, H2S=H2S, CO2=CO2, ignore_conflict=ignore_conflict)
        else:
            if ignore_conflict is False:
                self._check_conflicting_arguments(self.calc_Tr, 'Tr')
            self.Tr = Tr




